#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <sys/file.h>
#include <sys/types.h>
#include <fcntl.h>
#include <string.h>
#include <linux/cdrom.h>
#include "forms.h"

struct cdsector{
   char sync[12],
   header[4],
   subheader[8],
   data[2324],
   spare[4];
};

/**** Forms and Objects ****/

typedef enum {
        COM_STOP,
        COM_PLAY,
        COM_PREV,
        COM_REWIND,
        COM_FFWD,
        COM_NEXT,
        COM_MOVE
} COMMAND;

typedef struct{
	FL_FORM *rvcdform;
        FL_OBJECT *quit;
        FL_OBJECT *stop;
        FL_OBJECT *play;
        FL_OBJECT *prev;
        FL_OBJECT *next;
        FL_OBJECT *ffwd;
        FL_OBJECT *rewind;
        FL_OBJECT *msf;
        FL_OBJECT *track;
	FL_OBJECT *move;
        void *vdata;
        long ldata;
} FD_rvcdform;

FD_rvcdform *fd_rvcdform;
struct cdrom_tochdr tochdr;
struct cdrom_tocentry tocent[100];
int cdfile;
int stop=0;
int t = -1;		/* current track */
int lba0,lba1;	/* global lba range */
int t_lba0;		/* first lba (logical block address) of current track */
int t_lba;		/* current lba */
char *cd_device;
char m,s,f;
char m0,s0,f0;
char m1,s1,f1;
int track0=2;   /* first track in range; the MPEG stream usually begins here, track 2  */
int track1=0;   /* last tract in range, is usually the copyright or some kind of info */

extern FD_rvcdform * create_form_rvcdform(void);

static void title()
{
  fprintf(stderr,"xreadcdi version 0.4, Ales Makarov, 1997\n");
}

static void usage()
{
  fprintf(stderr,"usage: readcdi [-v] [-a M:S.F] [track[-[track]]]\n");
  fprintf(stderr,"options: -v verbose\n");
  fprintf(stderr,"\t -a M:S.F address of the starting sector in MSF format\n");
  fprintf(stderr,"\t track number of the starting/ending track\n\n");
}

static int msf_to_lba (char m, char s, char f)
{
  return (((m * CD_SECS) + s) * CD_FRAMES + f) - CD_BLOCK_OFFSET;
}

static void lba_to_msf (int lba, char *m, char *s, char *f)
{
        lba += CD_BLOCK_OFFSET;
        lba &= 0xffffff;  /* negative lbas use only 24 bits */
        *m = lba / (CD_SECS * CD_FRAMES);
        lba %= (CD_SECS * CD_FRAMES);
        *s = lba / CD_FRAMES;
        *f = lba % CD_FRAMES;
}

static int get_track_number(char m, char s, char f)
{
  int i=1;
  int lba,lba2;

  do
  {
    lba = msf_to_lba(m,s,f);
    lba2 = msf_to_lba(tocent[i].cdte_addr.msf.minute,tocent[i].cdte_addr.msf.second,tocent[i].cdte_addr.msf.frame);
    i++;
  } while ((lba>=lba2) && (i<=tochdr.cdth_trk1));

  return (i-1);
}

static void set_track_number(int track, char *m, char *s, char *f)
{
  *m = tocent[track-1].cdte_addr.msf.minute;
  *s = tocent[track-1].cdte_addr.msf.second;
  *f = tocent[track-1].cdte_addr.msf.frame;
}

static void next_sector()
{
  f++;
    if (f>=CD_FRAMES) {
       f = 0;
       s++;
       if (s>=CD_SECS) {
          s = 0;
          m++;
       }
    }
}

static void set_msf_label(char m, char s, char f)
{
  char buffer[9];

  sprintf(buffer,"%02d:%02d",m,s);
  fl_set_object_label(fd_rvcdform->msf, buffer);
}

static void set_track_label(int track)
{
  char buffer[6];

  sprintf(buffer,"%02d/%02d", track - track0 + 1, track1 - track0 + 1);
  fl_set_object_label(fd_rvcdform->track, buffer);
}

static void set_move_slider(char m, char s, char f)
{
  fl_set_slider_value(fd_rvcdform->move, msf_to_lba(m,s,f));
}

static void read_cdi(int verbose)
{
  int i;

  tochdr.cdth_trk0 = 1;
  tochdr.cdth_trk1 = 2;

  if (verbose) fprintf(stderr,"\nCD-i entries:\n");

  /* read individual tracks */
  i = tochdr.cdth_trk0;
    tocent[i-1].cdte_track = i;
    tocent[i-1].cdte_format = CDROM_MSF;
    tocent[i-1].cdte_addr.msf.minute = 0;
    tocent[i-1].cdte_addr.msf.second = 2;
    tocent[i-1].cdte_addr.msf.frame = 0;

    if (verbose) fprintf(stderr, "Track:%d start %02d:%02d.%02d\n",i, \
      tocent[i-1].cdte_addr.msf.minute, \
      tocent[i-1].cdte_addr.msf.second, \
      tocent[i-1].cdte_addr.msf.frame);

  i = tochdr.cdth_trk1;
    tocent[i-1].cdte_track = i;
    tocent[i-1].cdte_format = CDROM_MSF;
    tocent[i-1].cdte_addr.msf.minute = 1;
    tocent[i-1].cdte_addr.msf.second = 13;
    tocent[i-1].cdte_addr.msf.frame = 05;

    if (verbose) fprintf(stderr, "Track:%d start %02d:%02d.%02d\n",i, \
      tocent[i-1].cdte_addr.msf.minute, \
      tocent[i-1].cdte_addr.msf.second, \
      tocent[i-1].cdte_addr.msf.frame);

  /* read the lead-out track */
  tocent[tochdr.cdth_trk1].cdte_track = CDROM_LEADOUT;
  tocent[tochdr.cdth_trk1].cdte_format = CDROM_MSF;
  tocent[tochdr.cdth_trk1].cdte_addr.msf.minute = 56;
  tocent[tochdr.cdth_trk1].cdte_addr.msf.second = 0;
  tocent[tochdr.cdth_trk1].cdte_addr.msf.frame = 0;

  if (verbose) fprintf(stderr,"LeadOut start %02d:%02d.%02d\n\n", \
    tocent[tochdr.cdth_trk1].cdte_addr.msf.minute, \
    tocent[tochdr.cdth_trk1].cdte_addr.msf.second, \
    tocent[tochdr.cdth_trk1].cdte_addr.msf.frame);
}

int main(int argc, char *argv[])
{
    FL_OBJECT *obj;
    struct cdsector buf;
    struct cdrom_msf msf;
    int i,c,l;
    int x,y,z;
    int verbose=0;
    int sectaddr=0;
    u_int counter=0;

    cd_device = strdup("/dev/cdrom");

    /* open cdrom device */
    if ( (cdfile=open(cd_device, O_RDONLY)) == -1 ) {
        fprintf(stderr, "%s: can't open cdrom (%s)\n", argv[0], cd_device);
        exit(1);
    }

    /* parse commandline */
    for (i=1; i<argc; i++) {
        if (argv[i][0]=='-') {
            switch (argv[i][1]) {
                case 'v': verbose=1; break;
                case 'a': sectaddr=++i; break;
                default: title(); usage(); exit(1);
            }
        } else sscanf(argv[i],"%d-%d",&track0,&track1);
    }

    if (verbose) title();
    read_cdi(verbose);

    /* decide area */
    if (sectaddr) {
        track0=0;
        sscanf(argv[sectaddr],"%d:%d.%d",&x,&y,&z);
        m = (char) x; s = (char) y; f = (char) z;
    } else if ((track0>=tochdr.cdth_trk0) && (track0<=tochdr.cdth_trk1)) {
            set_track_number(track0,&m,&s,&f);
        } else {
            fprintf(stderr,"Bad track number (%d)! Available values <%d,%d>\n", 
               track0,tochdr.cdth_trk0,tochdr.cdth_trk1);
            exit(1);
        }

    m0 = m; s0 = s; f0 = f;

    if (!track1) track1=tochdr.cdth_trk1;
    if (track1<track0) track1=track0;
    if (track1>tochdr.cdth_trk1) {
        fprintf(stderr,"Bad track number (%d)! Available values <%d,%d>\n",
          track1,tochdr.cdth_trk0,tochdr.cdth_trk1);
            exit(1);
         }
    set_track_number(track1+1,&m1,&s1,&f1);

    lba0 = msf_to_lba(m0,s0,f0);
    lba1 = msf_to_lba(m1,s1,f1)-10;

    if (verbose) {
        fprintf(stderr,"Playing ");
        if (!sectaddr)
        {  
          if (track0!=track1) fprintf(stderr,"Track %d-%d ( ",track0,track1);
          else fprintf(stderr,"Track %d ( ",track0);
        }
        fprintf(stderr,"Area %02d:%02d.%02d - %02d:%02d.%02d",m0,s0,f0,m1,s1,f1);
        if (!sectaddr) fprintf(stderr," )");
        fprintf(stderr,"\n\n");
    }

  fl_initialize(&argc, argv, "FormDemo", 0, 0);
  fd_rvcdform = create_form_rvcdform();

  fl_set_form_dblbuffer(fd_rvcdform->rvcdform, 1);
  fl_show_form(fd_rvcdform->rvcdform, FL_PLACE_FREE, FL_FULLBORDER, "ReadCD-i");

  do {
    l = msf_to_lba(m,s,f);
    t_lba = l;
    if ((l<lba1) && (l>=lba0)) {
      		int new_t = get_track_number(m,s,f);
      		if (new_t != t) {
      			t = new_t;
      			t_lba0 = l;
      		}

      if (counter % 16 == 0) {
        obj = fl_check_forms();
      }
      if (counter % 64 == 0) {
        set_track_label(t);
        set_msf_label(m,s,f);
        set_move_slider(m,s,f);
      }

      counter++;

      msf.cdmsf_min0 = m;
      msf.cdmsf_sec0 = s;
      msf.cdmsf_frame0 = f;

      if (!stop) {
        memcpy (&buf, &msf, sizeof (msf));
        if ( ioctl(cdfile, CDROMREADRAW, &buf) == -1 ) {
          fprintf(stderr, "%s: ioctl cdromreadraw\n", argv[0]);
          exit(1);
        }

/* audio & video only */
        if ((buf.subheader[1]==1) && (((buf.subheader[2]==0x62) &&
            (buf.subheader[3]==0x0f)) || ((buf.subheader[2]==0x64) &&
            (buf.subheader[3]==0x7f))))
          write(1, buf.data, 2324);
        next_sector();
      }
    } else {
       stop = 1;
       fl_set_object_lcol(fd_rvcdform->stop,FL_GREEN);
       fl_set_object_lcol(fd_rvcdform->play,FL_BLACK);
       set_track_number(track0,&m,&s,&f);
       set_track_label(track0);
       set_msf_label(m,s,f);
       set_move_slider(m,s,f);
    }
  } while (obj != fd_rvcdform->quit);

  close(cdfile);
  return 0;
}

void
command(fl_object, param)
        FL_OBJECT *fl_object;
        long param;
{
        COMMAND com = param;
        
        switch (com) {
                case COM_STOP:
			stop = 1;
                        fl_set_object_lcol(fl_object,FL_GREEN);
                        fl_set_object_lcol(fd_rvcdform->play,FL_BLACK);
                        break;
                case COM_PLAY:
			stop = 0;
                        fl_set_object_lcol(fl_object,FL_GREEN);
                        fl_set_object_lcol(fd_rvcdform->stop,FL_BLACK);
                        break;
                case COM_PREV:
                	if (t_lba < t_lba0 + 75) {
                		t--;
                	}
	                if (t>=track0) set_track_number(t,&m,&s,&f);
	                set_track_label(get_track_number(m,s,f));
		        set_msf_label(m,s,f);
                        set_move_slider(m,s,f);
                        break;
                case COM_NEXT:
	                if (t<track1) set_track_number(t+1,&m,&s,&f);
                        set_track_label(get_track_number(m,s,f));
                        set_msf_label(m,s,f);
                        set_move_slider(m,s,f);
                        break;
                case COM_REWIND:
	 		if (m>1) m--;
                        set_track_label(get_track_number(m,s,f));
                        set_msf_label(m,s,f);
                        set_move_slider(m,s,f);
                        break;
                case COM_FFWD:
			if (m+1<tocent[tochdr.cdth_trk1].cdte_addr.msf.minute) m++;
                        set_track_label(get_track_number(m,s,f));
                        set_msf_label(m,s,f);
                        set_move_slider(m,s,f);
                        break;
                case COM_MOVE:
			lba_to_msf((int) fl_get_slider_value(fl_object),&m,&s,&f);
                        set_track_label(get_track_number(m,s,f));
                        set_msf_label(m,s,f);
                        break;
                default:
                        break;
        }
}

FD_rvcdform *create_form_rvcdform(void)
{
  FL_OBJECT *obj;
  FD_rvcdform *fdui = (FD_rvcdform *) fl_calloc(1, sizeof(FD_rvcdform));
  int old_bw = fl_get_border_width();

  fl_set_border_width(-1);
  fdui->rvcdform = fl_bgn_form(FL_UP_BOX,270,50);
    fdui->quit = obj = fl_add_button(FL_NORMAL_BUTTON,5,5,34,20,"Quit");
      fl_set_object_lstyle(obj,FL_BOLD_STYLE);
/*      fl_set_idle_callback(command,COM_STOP); */
    fdui->stop = obj = fl_add_button(FL_NORMAL_BUTTON,43,5,21,20,"||");
      fl_set_object_lstyle(obj,FL_BOLD_STYLE);
      fl_set_object_lcol(obj,FL_BLACK);
      fl_set_object_callback(obj,command,COM_STOP);
    fdui->play = obj = fl_add_button(FL_NORMAL_BUTTON,65,5,21,20,"@>");
      fl_set_object_lcol(obj,FL_GREEN);
      fl_set_object_callback(obj,command,COM_PLAY);
    fdui->prev = obj = fl_add_button(FL_NORMAL_BUTTON,90,5,21,20,"@4->|");
      fl_set_object_callback(obj,command,COM_PREV);
    fdui->rewind = obj = fl_add_button(FL_NORMAL_BUTTON,112,5,21,20,"@<<");
      fl_set_object_callback(obj,command,COM_REWIND);
    fdui->ffwd = obj = fl_add_button(FL_NORMAL_BUTTON,134,5,21,20,"@>>");
      fl_set_object_callback(obj,command,COM_FFWD);
    fdui->next = obj = fl_add_button(FL_NORMAL_BUTTON,156,5,21,20,"@->|");
      fl_set_object_callback(obj,command,COM_NEXT);

    obj = fl_add_text(FL_NORMAL_TEXT,184,3,40,25,"mm:ss");
      fl_set_object_lalign(obj,FL_ALIGN_CENTER|FL_ALIGN_INSIDE);

    obj = fl_add_text(FL_NORMAL_TEXT,230,3,30,25,"Track");
      fl_set_object_lalign(obj,FL_ALIGN_LEFT|FL_ALIGN_INSIDE);

    fdui->move = obj = fl_add_slider(FL_HOR_NICE_SLIDER,5,30,172,15,"");
      fl_set_object_boxtype(obj,FL_FLAT_BOX);
      fl_set_object_lsize(obj,FL_DEFAULT_SIZE);
      fl_set_object_lalign(obj,FL_ALIGN_LEFT);
      fl_set_object_callback(obj,command,COM_MOVE);
      fl_set_slider_return(obj, FL_RETURN_CHANGED);
/*      fl_set_slider_size(obj, 0.05); */
      fl_set_slider_bounds(obj, lba0, lba1);
      fl_set_slider_value(obj, lba0);

    fdui->msf = obj = fl_add_text(FL_NORMAL_TEXT,183,30,42,14,"00:00");
      fl_set_object_boxtype(obj,FL_DOWN_BOX);
      fl_set_object_color(obj,FL_TOP_BCOL,FL_MCOL);
      fl_set_object_lalign(obj,FL_ALIGN_CENTER|FL_ALIGN_INSIDE);
      fl_set_object_lstyle(obj,FL_TIMES_STYLE);

    fdui->track = obj = fl_add_text(FL_NORMAL_TEXT,232,30,32,14,"00/00");
      fl_set_object_boxtype(obj,FL_DOWN_BOX);
      fl_set_object_color(obj,FL_TOP_BCOL,FL_MCOL);
      fl_set_object_lalign(obj,FL_ALIGN_RIGHT|FL_ALIGN_INSIDE);
      fl_set_object_lstyle(obj,FL_TIMES_STYLE);

  fl_end_form();
  fl_set_border_width(old_bw);

  return fdui;
}
